## Distribution class for variable development models

setClass("VD.dist",
         representation(params = "list",
                        distribution.functions = "list",
                        deriv.functions = "list"))

setMethod("initialize", "VD.dist",
          function(.Object, distribution.name, params = NULL, pdf.derivs = NULL, cdf.derivs = NULL, quantile.derivs = NULL) {
            
            if(class(distribution.name) != "character") stop("Error, in initialize for VD.dist, distribution.name should be a character string.");
            
            .Object@params <- c(list(1), params); 
            calc.names <- c("pdf","cdf", "random", "quantile");
            calc.prefixes <- c("d","p","r","q");
            function.names <- paste(calc.prefixes, distribution.name, sep="");
            for(i in seq(1, along = calc.names)) {
              .Object@distribution.functions[[calc.names[i]]] <- eval.parent(as.name(function.names[i]));
            }
            
            for(deriv.type in c("pdf", "cdf", "quantile")) {
              deriv.name <- paste(deriv.type, ".derivs", sep="");
              this.derivs <- eval(substitute(dn, list(dn = as.name(deriv.name))));
              if(!is.null(this.derivs)) {
                .Object@deriv.functions[[deriv.type]] <- this.derivs
              }
            }
            .Object
          })

setMethod("plot",
          signature(y = "VD.dist"),
          function(x, y, lower.quantile = 0, upper.quantile = 0.9999, cdf = FALSE, ...) {
            browser();
            if(missing(x)) {
              x.range <- ceiling(VD.dist.calc(c(lower.quantile, upper.quantile), y, "quantile"));
              x <- x.range[1]:x.range[2];
            }
            probability <- VD.dist.calc(x, y, if(cdf) "cdf" else "pdf");
            callGeneric(x, probability, ...); 
          }
          )

setMethod("plot",
          signature(x = "VD.dist"),
          function(x, y, ...) {
            browser();
            if(!missing(y)) writeLines("Warning: y is ignored");
            callGeneric(y = x, ...); 
          }
          )


VD.dist <- function(distribution.name, params = NULL, pdf.derivs = NULL, cdf.derivs = NULL, quantile.derivs = NULL) {
  new("VD.dist", distribution.name = distribution.name, params = params, pdf.derivs = pdf.derivs, cdf.derivs = cdf.derivs, quantile.derivs = quantile.derivs)
}

VD.dist.calc <- function(x, distr, calc, make.discrete = FALSE, new.params = NULL) {
  if(class(distr) != "VD.dist") {stop("error, must supply an object of class VD.dist");}
  do.new.params <- !is.null(new.params);
  if(!make.discrete) {
    updated.params <- distr@params;
    updated.params[[1]] <- as.name("x");
    if(!do.new.params) {
      return(do.call(distr@distribution.functions[[calc]], updated.params))
    } 
    replace.indices <- match(names(new.params), names(updated.params), 0);
    updated.params[replace.indices] <- new.params[replace.indices > 0];
    other.params <- new.params[replace.indices == 0];
    lnc <- length(updated.params);
    if(length(other.params) > 0) {
      for(i in 1:length(other.params)) {
        updated.params[[lnc + i]] <- other.params[[i]];
      }
      names(updated.params)[lnc + 1:length(other.params)] <- names(other.params);
    }
    return(do.call(distr@distribution.functions[[calc]], updated.params))
  } else {
    if(!(calc %in% c("pdf", "cdf"))) {stop("Error: make.discrete only makes sense for calc = \"pdf\" or \"cdf\" ");}
    updated.params <- distr@params;
    if(calc == "pdf") {
      ceil.x <- ceiling(x);
      range.ceil.x <- range(ceil.x);
      seq.x <- (range.ceil.x[1]-1):range.ceil.x[2];
      updated.params[[1]] <- as.name("seq.x");
    } else {
      ceil.x <- ceiling(x);
      updated.params[[1]] <- as.name("ceil.x");
    }
    if(do.new.params) {
      replace.indices <- match(names(new.params), names(updated.params), 0);
      updated.params[replace.indices] <- new.params[replace.indices > 0]
      other.params <- new.params[replace.indices == 0];
      lnc <- length(updated.params)
      if(length(other.params) > 0) {
        for(i in 1:length(other.params)) {
          updated.params[[lnc + i]] <- other.params[[i]]
        }
        names(updated.params)[lnc + 1:length(other.params)] <- names(other.params);
      }
    }
    if(calc == "pdf") {
      pdf.vals <- diff(do.call(distr@distribution.functions[["cdf"]], updated.params));
      return(pdf.vals[ceil.x - range.ceil.x[1] + 1])
    } else {
      return(do.call(distr@distribution.functions[[calc]], updated.params));
    }
  }
}

VD.dist.deriv <- function(x, distr, calc, param, force.numerical = FALSE, delta = 0.0001, make.discrete = FALSE, relative = FALSE, new.params = NULL) {
  if(class(distr) != "VD.dist") {stop("error, must supply an object of class VD.dist");}
  if(calc == "random") {stop("error, does not make sense to take derivatives of the random generator");}
  do.new.params <- !is.null(new.params);
  do.numerical <- TRUE
  if(!force.numerical) {
    do.numerical <- !(param %in% names(distr@deriv.functions[[calc]]))
  }
  if(!do.numerical) {
    if(make.discrete) {writeLines("Warning: make.discrete = TRUE has no impact since you are calling a supplied derivative function.")}
    updated.params <- distr@params;
    updated.params[[1]] <- as.name("x");
    if(do.new.params) {
      replace.indices <- match(names(new.params), names(updated.params), 0);
      updated.params[replace.indices] <- new.params[replace.indices > 0] 
      other.params <- new.params[replace.indices == 0];
      lnc <- length(updated.params)
      if(length(other.params) > 0) {
        for(i in 1:length(other.params)) {
          updated.params[[lnc + i]] <- other.params[[i]]
        }
        names(updated.params)[lnc + 1:length(other.params)] <- names(other.params);
      }
    }

    if(!relative) {
      return(do.call(distr@deriv.functions[[calc]][[param]], updated.params));
    } else {
      return(do.call(distr@deriv.functions[[calc]][[param]], updated.params) / VD.dist.calc(x, distr, calc, make.discrete, new.params = new.params));
    }
  } else {
    vals <- VD.dist.calc(x, distr, calc, make.discrete, new.params = new.params);
    param.plus.delta <- distr@params[[param]] + delta
    new.new.params <- list()
    new.new.params[[param]] <- param.plus.delta;
    new.params <- c(new.new.params, new.params);
    plus.delta.vals <-VD.dist.calc(x, distr, calc, make.discrete, new.params = new.params);

    if(!relative) {
      return( ( plus.delta.vals - vals ) / delta );
    } else {
      return( ( ( plus.delta.vals / vals ) - 1 ) / delta )
    }
  }
}


