## Code for demographic analysis of variable development models using
## Monte Carlo integration of the Euler equation


VD.fecundity.constant <- function(age, age.maturity, F) {
  return(rep(F, length(age.maturity)));
}
  
dinfinite <- function(x) {rep(NA, length(x));}
qinfinite<- function(p,...) {rep(NA, length(p));}
pinfinite <- function(x) {rep(1, length(x));}
rinfinite <- function(n) {rep(Inf, n);}

dgeomp1 <- function(x, prob) {dgeom(x-1, prob);}
qgeomp1 <- function(p, prob, ...) {qgeom(p, prob, ...) + 1;}
rgeomp1 <- function(n, prob) {rgeom(n, prob) + 1;}
pgeomp1 <- function(x, prob, ...) {pgeom(x-1, prob, ...);}


VD.run <- function(VDM, stages = NULL, verbose = FALSE) {
  ans <- VD.sim(VDM);
  if(is.null(stages)) stages <- 1:num.stages(VDM);
  for(iS in stages) {
    ans <- VD.run.stage(VDM, ans, iS, verbose = verbose);
  }
  ans
}

VD.calc.conditional.info <- function(VDM, VDS, iS) {
  
  if(iS < 1 | iS >= num.stages(VDM)) {stop(paste("Error: VD.get.conditional.info got stage", iS, sep = ""));}
  ans<-list()
  
  Sig11 <- gauss.cov(VDM)[1:iS, 1:iS]
  Sig21 <- matrix(gauss.cov(VDM)[ iS+1, 1:iS], nrow = 1, ncol = iS)
  Sig22 <- gauss.cov(VDM)[(iS + 1), (iS + 1)]
  invSig11 <- solve(Sig11);
  dev.table.Z <- compile.dev.table(VDS, include.dev = FALSE, include.Z = TRUE, survivors = TRUE)$dev.table.Z;
  
  ans$conditional.means <- as.vector(as.matrix(dev.table.Z) %*% t(Sig21 %*% invSig11));
  ans$conditional.var <- Sig22 - Sig21 %*% invSig11 %*% t(Sig21)
  ans$conditional.sd <- sqrt(ans$conditional.var)
  ans$num.alive <- length(ans$conditional.means)
  ans
}

VD.run.stage <- function(VDM, VDS, iS, verbose = FALSE) {
  indep <- stage.independent(VDM)[iS];
  all.indep <- all.independent(VDM);
  first <- iS == 1
  iSm1 <- iS-1;
  VD.controls.iS <- controls(VDM)[iS,]
  batch.size <- VD.controls.iS$batch.size;
  alive.target <- VD.controls.iS$alive.target;
  dead.target <- VD.controls.iS$dead.target;
  max.dead <- VD.controls.iS$max.dead;
  max.total <- VD.controls.iS$max.total;
  ragged.save <- VD.controls.iS$ragged.save;
  total.count <- alive.count <- dead.count <- 0;
  all.alive.dev.time <- all.alive.dev.time.contin <- all.dead.death.time <- numeric(0)
  if(!all.indep) all.alive.Z<-numeric(0)
  if(!first) all.alive.source.rows <- all.dead.source.rows <- numeric(0)
  if(!indep) {
    conditionalInfo <- VD.calc.conditional.info(VDM, VDS, iSm1);
    conditional.sd <- conditionalInfo$conditional.sd
  }
  
  if(!first) {
    num.prev.alive <- stageMC(VDS)[[iSm1]]$num.alive;
    source.rows <- 1:num.prev.alive
    i.prev <- 1;
  }
  
  done <- alive.done <- dead.done <- dead.done.saving <- FALSE;
  alive.discarded <- dead.discarded <- 0
  if(dead.target == 0) dead.done <- TRUE;
  if(max.dead == 0) dead.done.saving <- TRUE
  while(!done) {
    
    if(!first) this.batch.source.rows <- numeric(0);    
    if(!indep) this.batch.conditional.means <- numeric(0);
    if(!first | !indep) {
      num.taken <- 0;
      while(num.taken < batch.size) {
        next.end <- min(num.prev.alive, i.prev + batch.size - num.taken - 1);
        if(!indep) this.batch.conditional.means <- c(this.batch.conditional.means,
                                                     conditionalInfo$conditional.means[i.prev:next.end])
        if(!first) this.batch.source.rows <- c(this.batch.source.rows, source.rows[i.prev:next.end]);
        num.taken <- num.taken + (next.end - i.prev + 1);
        i.prev <- if(next.end == num.prev.alive) 1 else next.end + 1;
      }
    }
    if(!indep) {
      this.batch.conditional.normal <- rnorm(batch.size, this.batch.conditional.means, conditional.sd)
     } else {
      this.batch.conditional.normal <- rnorm(batch.size, 0, 1)
    }
    this.batch.marginal.cumulative <- pnorm(this.batch.conditional.normal)
    this.batch.dev.time.contin <- VD.dist.calc(this.batch.marginal.cumulative, marginal.durations(VDM)[[iS]], "quantile");

    this.batch.dev.time <- ceiling(this.batch.dev.time.contin)
    this.batch.death.time <- ceiling(VD.dist.calc(batch.size, marginal.death.times(VDM)[[iS]], "random"));
    this.batch.survival <- this.batch.death.time > this.batch.dev.time
    this.batch.num.survived <- sum(this.batch.survival)

    if(alive.count + this.batch.num.survived >= alive.target) {
      if(ragged.save) {
        all.alive.dev.time <- c(all.alive.dev.time, this.batch.dev.time[this.batch.survival])
        all.alive.dev.time.contin <- c(all.alive.dev.time.contin, this.batch.dev.time.contin[this.batch.survival])
        if(!all.indep) all.alive.Z <- c(all.alive.Z, this.batch.conditional.normal[this.batch.survival])
        if(!first) all.alive.source.rows <- c(all.alive.source.rows, this.batch.source.rows[this.batch.survival])
        alive.count <- alive.count + this.batch.num.survived;
        alive.discarded <- 0;
      } else {
        if(!alive.done) {
          all.alive.dev.time <- c(all.alive.dev.time, this.batch.dev.time[this.batch.survival][1:(alive.target - alive.count)])
          all.alive.dev.time.contin <- c(all.alive.dev.time.contin, this.batch.dev.time.contin[1:(alive.target - alive.count)])
          if(!all.indep) all.alive.Z <- c(all.alive.Z, this.batch.conditional.normal[this.batch.survival][1:(alive.target - alive.count)])
          if(!first) all.alive.source.rows <- c(all.alive.source.rows, this.batch.source.rows[this.batch.survival][1:(alive.target - alive.count)])
        }
        alive.discarded <- alive.discarded + alive.count + this.batch.num.survived - alive.target
        alive.count <- alive.target;        
      }
      alive.done <- TRUE
    } else {
      all.alive.dev.time <- c(all.alive.dev.time, this.batch.dev.time[this.batch.survival])
      all.alive.dev.time.contin <- c(all.alive.dev.time.contin, this.batch.dev.time.contin[this.batch.survival])
      if(!all.indep) all.alive.Z <- c(all.alive.Z, this.batch.conditional.normal[this.batch.survival])
      if(!first) all.alive.source.rows <- c(all.alive.source.rows, this.batch.source.rows[this.batch.survival])
      alive.count <- alive.count + this.batch.num.survived;
    }
    
    this.batch.num.died <- batch.size - this.batch.num.survived
    if(dead.count + this.batch.num.died >= max.dead) {
      if(max.dead - dead.count >= 1) {
        all.dead.death.time <- c(all.dead.death.time, this.batch.death.time[!this.batch.survival][1:(max.dead - dead.count)])
        if(!first) all.dead.source.rows <- c(all.dead.source.rows, this.batch.source.rows[!this.batch.survival][1:(max.dead - dead.count)])
      }
      dead.discarded <- dead.discarded + dead.count + this.batch.num.died - max.dead      
      dead.count <- max.dead;
    } else {
      all.dead.death.time <- c(all.dead.death.time, this.batch.death.time[!this.batch.survival])
      if(!first) all.dead.source.rows <- c(all.dead.source.rows, this.batch.source.rows[!this.batch.survival])
      dead.count <- dead.count + this.batch.num.died
    }
    
    if(dead.count + dead.discarded > dead.target) dead.done <- TRUE;
    if(dead.done & alive.done) done <- TRUE
    
    total.count <- total.count + batch.size;
    if(verbose) {
      writeLines(paste("New Batch ", batch.size, this.batch.num.survived, this.batch.num.died))
      writeLines(paste("alive.count", alive.count, length(all.alive.dev.time), ". alive.discared ", alive.discarded, ". dead.count", dead.count, length(all.dead.death.time), ". dead.discarded", dead.discarded, ". total.count", total.count, ". ", alive.done, dead.done, done)) 
    }
    
    if(total.count >= max.total) done <- TRUE
  }

  simulation.summary(VDS)[iS,c("Stage", "prob.surv.stage", "total.count", "total.alive", "total.dead", "discarded.alive", "discarded.dead")] <-
    c(iS, (alive.count + alive.discarded) / total.count, total.count, alive.count, dead.count, alive.discarded, dead.discarded)

  stageMC(VDS)[[iS]] <- 
  if(!first) {
    if(!all.indep) {
      (list(
                  num.alive = alive.count,
                  aliveT = structure(data.frame(all.alive.dev.time, all.alive.dev.time.contin, all.alive.Z, all.alive.source.rows),
                    names = c("dev.time", "dev.time.contin", "Z", "SourceRow")),
                  deadT = structure(data.frame(all.dead.death.time, all.dead.source.rows), names = c("DeathTime", "SourceRow"))))
    } else {
      (list(num.alive = alive.count,
                  aliveT = structure(data.frame(all.alive.dev.time, all.alive.dev.time.contin, all.alive.source.rows),
                    names = c("dev.time", "dev.time.contin", "SourceRow")),
                  deadT = structure(data.frame(all.dead.death.time, all.dead.source.rows), names = c("DeathTime", "SourceRow"))))
      
    }
  } else {
    if(!all.indep) {
      (list(
                  num.alive = alive.count,
                  aliveT = structure(data.frame(all.alive.dev.time, all.alive.dev.time.contin, all.alive.Z), names = c("dev.time", "dev.time.contin", "Z")),
                  deadT = structure(data.frame(all.dead.death.time), names = c("DeathTime"))))
    } else {
      (list(num.alive = alive.count,
                  aliveT = structure(data.frame(all.alive.dev.time, all.alive.dev.time.contin), names = c("dev.time", "dev.time.contin")),

            deadT = structure(data.frame(all.dead.death.time), names = c("DeathTime"))))
    }
  }
  VDS
}

compile.dev.table <- function(VDS, final.stage = NULL, init.stage = 1, total.age = TRUE, include.dev = TRUE, include.contin = FALSE, include.Z = FALSE, survivors = FALSE) {

  num.stages <- num.stages(VDS);
  if(is.null(final.stage)) final.stage <- length(stageMC(VDS));
  if(final.stage > num.stages | init.stage < 1 | final.stage < init.stage) stop(paste("Error in stage request from ", init.stage, " to ", final.stage,sep=""))
  end.dead <- final.stage < num.stages & !survivors

  requestedStages <- final.stage - init.stage + 1;
  
  m <- if(!end.dead) {
    stageMC(VDS)[[final.stage]]$num.alive
  } else {
    simulation.summary(VDS)[final.stage, "total.dead"];
  }
  dev.table <- structure(data.frame(matrix(nrow = m, ncol = requestedStages)), names = stage.names(VDS)[init.stage:final.stage])
  if(end.dead) {
    if(include.contin | include.Z) {writeLines(paste("warning: not including contin or Z in compile.dev.table because the last stage has not been requested.", sep = ""));}
    include.contin <- include.Z <- FALSE;
  }
  if(include.contin) dev.table.contin <- dev.table;
  if(include.Z) dev.table.Z <- dev.table;
  
  survival.weight <- 1;
  total.sim.kept <- 0;
    
  indexStage <- requestedStages;
  if(!end.dead) {
    if(include.dev) dev.table[ , indexStage] <- stageMC(VDS)[[final.stage]]$aliveT$dev.time;
    if(include.contin) dev.table.contin[ , indexStage] <- stageMC(VDS)[[final.stage]]$aliveT$dev.time.contin;
    if(include.Z) dev.table.Z[, indexStage] <- stageMC(VDS)[[final.stage]]$aliveT$Z;
    SourceRow <- stageMC(VDS)[[final.stage]]$aliveT$SourceRow
    final.sim.kept <- simulation.summary(VDS)[final.stage, "total.alive"]
  } else {
    if(include.dev) dev.table[ , indexStage] <- stageMC(VDS)[[final.stage]]$deadT$DeathTime;
    SourceRow <- stageMC(VDS)[[final.stage]]$deadT$SourceRow
    survival.weight <- survival.weight * (1-simulation.summary(VDS)[final.stage, "prob.surv.stage"])
    final.sim.kept <- simulation.summary(VDS)[final.stage, "total.dead"]
  }
  if(requestedStages > 1) {
    for(thisStage in seq(final.stage - 1, init.stage, by = -1)) {
      indexStage <- indexStage - 1;
      if(include.dev) dev.table[ , indexStage] <- stageMC(VDS)[[thisStage]]$aliveT$dev.time[SourceRow];
      if(include.contin) dev.table.contin[, indexStage] <- stageMC(VDS)[[thisStage]]$aliveT$dev.time.contin[SourceRow];
      if(include.Z) dev.table.Z[, indexStage] <- stageMC(VDS)[[thisStage]]$aliveT$Z[SourceRow];
      survival.weight <- survival.weight * simulation.summary(VDS)[thisStage, "prob.surv.stage"];
      if(thisStage > 1) SourceRow <- stageMC(VDS)[[thisStage]]$aliveT$SourceRow[SourceRow];
    }

    if(total.age & include.dev) {
      if(requestedStages > 1) {
        for(indexStage in 2:requestedStages) {
          dev.table[, indexStage] <- dev.table[, indexStage - 1] + dev.table[, indexStage]
        }
      }
    }
  }

  total.sim.kept <- sum(simulation.summary(VDS)$total.dead[-num.stages]) + simulation.summary(VDS)$total.alive[num.stages]
  
  stage.kept.weight <-  total.sim.kept / final.sim.kept; 

  list(dev.table = if(include.dev) dev.table else NULL,
       dev.table.contin = if(include.contin) dev.table.contin else NULL,
       dev.table.Z = if(include.Z) dev.table.Z else NULL,
       survival.weight = survival.weight,
       stage.kept.weight = stage.kept.weight,
       shared.weight = survival.weight * stage.kept.weight,
       MC.sample.size.final.stage = final.sim.kept,
       MC.sample.size.all = total.sim.kept
       )
}

euler <- function(r, fecundity.by.age) {
  sum(fecundity.by.age$mean.fecundity * exp(-r*fecundity.by.age$age))
}

VD.deriv.euler.dr <- function(r, fecundity.by.age) {
  -sum(fecundity.by.age$mean.fecundity * fecundity.by.age$age * exp(-r*fecundity.by.age$age))
}

VD.solve.euler <- function(fecundity.by.age, ...) {
  
  z <- fecundity.by.age$age;
  mean.fecundity <- fecundity.by.age$mean.fecundity;
  euler.m1 <- function(r) {
    sum(mean.fecundity * exp(-r*z)) - 1
  }
  lower <- -50/max(z);
  ans <- uniroot(euler.m1, interval = c(lower,3), ...)
  
  ans$root
}

calc.average.surv.rep.by.age <- function(dev.table, F = NULL, weights = NULL, use.shared.weight = TRUE, deriv.param = NULL) {
  if(is(F, "VD.model")) {    
    ans <- calc.average.surv.rep.by.age.funcM(dev.table, F@fecundity, weights = weights, use.shared.weight = use.shared.weight, deriv.param = deriv.param); 
  } else {
    if(is.numeric(F)) {
      F #force evaluation.
      fec <- VD.fec(VD.fecundity.constant, list(F = F));
      ans <- calc.average.surv.rep.by.age.funcM(dev.table, F = fec, weights = weights, use.shared.weight = use.shared.weight, deriv.param = deriv.param)
    } else {
      ans <- calc.average.surv.rep.by.age.funcM(dev.table, F = F, weights = weights, use.shared.weight = use.shared.weight, deriv.param = deriv.param)
    }
  }
  ans
}


calc.average.surv.rep.by.age.funcM <- function(dev.table, F, weights = NULL, use.shared.weight = TRUE, deriv.param = NULL) {
  colDie <- dim(dev.table$dev.table)[2];
  colMature <- colDie - 1;
  ageDie <- dev.table$dev.table[,colDie];
  if(colMature > 0 ){
    ageMature <- dev.table$dev.table[,colMature];
  } else {
    ageMature <- rep(0, length(ageDie));
  }
  zMin <- min(ageMature);
  zMax <- max(ageDie);
  numZ <- zMax - zMin + 1;
  ans <- data.frame(age = zMin:zMax, mean.fecundity = rep(NA, numZ))
  z <- zMin;
  anscol <- rep(0, numZ);
  if(is.null(deriv.param)) {
    if(is.null(weights)) {
      for(iz in 1:numZ) {
        anscol[iz] <- sum(VD.fec.calc(age = z, age.maturity = ageMature[ageMature < z & ageDie >= z], F));
        z <- z+ 1;
      }
      ans[,2] <- anscol;
    } else {
      for(iz in 1:numZ) {
        bool <- ageMature < z & ageDie >= z
        anscol[iz] <-sum(VD.fec.calc(age = z, age.maturity = ageMature[bool], F ) * weights[bool]);
        z <- z + 1;
      }
      ans[,2] <- anscol;
    }
  } else {
    if(is.null(weights)) {
      for(iz in 1:numZ) {                                        
        anscol[iz] <- sum(VD.fec.deriv(age = z, age.maturity = ageMature[ageMature < z & ageDie >= z], F, param = deriv.param));
        z <- z+ 1;
      }
      ans[,2] <- anscol;
    } else {
      for(iz in 1:numZ) {
        bool <- ageMature < z & ageDie >= z
        anscol[iz] <- sum(VD.fec.deriv(age = z, age.maturity = ageMature[bool], F ) * weights[bool], param = deriv.param);
        z <- z + 1;
      }
      ans[,2] <- anscol;
    }

  }
  if(use.shared.weight) {
    ans[,2] <- ans[,2] * dev.table$shared.weight / dev.table$MC.sample.size.all
  } else {
    ans[,2] <- ans[,2] / dev.table$MC.sample.size.all
  }
  ans
}


make.SASD <- function(VDS, r) {
  num.stages <- num.stages(VDS);
  final.stage <- num.stages;
  dev.tables.alive <- list();
  dev.tables.dead <- list();
  for(iS in 1:final.stage) {
    DT <- compile.dev.table(VDS, iS);
    dev.tables.dead[[iS]] <- DT$dev.table
  }
  if(final.stage > 1) {
    for(iS in 1:(final.stage - 1)) {
      DT <- compile.dev.table(VDS, iS, survivors = TRUE);
      dev.tables.alive[[iS]] <- DT$dev.table;
    }
  }
  max.age <- max(dev.tables.dead[[final.stage]][, final.stage]);
  age.vector <- 1:max.age;
  num.ages <- length(age.vector);
  dying.ASD <- maturing.ASD <- SASD <- matrix(0, ncol = final.stage, nrow = num.ages);

  prob.surv.stage <- simulation.summary(VDS)$prob.surv.stage;
  nj.alive <- simulation.summary(VDS)$total.alive;
  nj.dead <- simulation.summary(VDS)$total.dead;
  
  for(iS in 1:final.stage) {
    if(iS > 1) {
      if(iS < final.stage) {
        nT <- nj.alive[iS] + nj.dead[iS];
        weight.dead <- prod(prob.surv.stage[1:(iS - 1)]) * (1-prob.surv.stage[iS]) / (nj.dead[iS] / nT);
        weight.alive <- prod(prob.surv.stage[1:iS]) / (nj.alive[iS] / nT);
        this.age.range <- c(min(dev.tables.dead[[iS]][, iS-1], dev.tables.alive[[iS]][, iS - 1]),
                            max(dev.tables.dead[[iS]][, iS], dev.tables.alive[[iS]][, iS]))
        for(a in seq(this.age.range[1], this.age.range[2])) {
          ai <- a;
          SASD[ai, iS] <- (weight.dead * sum(dev.tables.dead[[iS]][, iS] >= a & dev.tables.dead[[iS]][, iS-1] < a) +
            weight.alive * sum(dev.tables.alive[[iS]][, iS] >= a & dev.tables.alive[[iS]][, iS-1] < a))/nT;
          maturing.ASD[ai, iS] <- weight.alive * sum(dev.tables.alive[[iS]][, iS] == a) / nT;
          dying.ASD[ai, iS] <- weight.dead * sum(dev.tables.dead[[iS]][, iS] == a) / nT;
        }
      } else { # iS == final.stage
        nT <- nj.alive[iS]; ## these are "matured" from final stage, i.e. dead
        weight.dead <- prod(prob.surv.stage[1:(iS - 1)]) * 1;
        this.age.range <- c(min(dev.tables.dead[[iS]][, iS-1]),
                            max(dev.tables.dead[[iS]][, iS]));
        for(a in seq(this.age.range[1], this.age.range[2])) {
          ai <- a ;
          SASD[ai, iS] <- weight.dead * sum(dev.tables.dead[[iS]][, iS] >= a & dev.tables.dead[[iS]][, iS-1] < a) / nT;
          dying.ASD[ai, iS] <- weight.dead * sum(dev.tables.dead[[iS]][, iS] == a) / nT;
        }
      }
    } else { ## iS == 1
      if(iS < final.stage) {
        nT <- nj.alive[iS] + nj.dead[iS];
        weight.dead <- (1-prob.surv.stage[iS]) / (nj.dead[iS] / nT);
        weight.alive <- (prob.surv.stage[iS]) / (nj.alive[iS] / nT);

        this.age.range <- c(0,
                            max(dev.tables.dead[[iS]][, iS], dev.tables.alive[[iS]][, iS]))
        for(a in seq(this.age.range[1], this.age.range[2])) {
          ai <- a;
          SASD[ai, iS] <- (weight.dead * sum(dev.tables.dead[[iS]][, iS] >= a) +
            weight.alive * sum(dev.tables.alive[[iS]][, iS] >= a)) / nT;
          maturing.ASD[ai, iS] <- weight.alive * sum(dev.tables.alive[[iS]][, iS] == a) / nT;
          dying.ASD[ai, iS] <- weight.dead * sum(dev.tables.dead[[iS]][, iS] == a) / nT;
        }
      } else { # iS == final.stage
        nT <- nj.dead[iS];
        this.age.range <- c(0,
                            max(dev.tables.dead[[iS]][, iS]));
        for(a in seq(this.age.range[1], this.age.range[2])) {
          ai <- a;
          SASD[ai, iS] <- sum(dev.tables.dead[[iS]][, iS] >= a) / nT;
          dying.ASD[ai, iS] <- sum(dev.tables.dead[[iS]][, iS] == a) / nT;
        }
      }
    }
  }
  # Do tallies.
  discounts <- exp(-r * (age.vector - 1));
  for(iS in 1:final.stage) {
    SASD[, iS] <- SASD[, iS] * discounts;
    dying.ASD[, iS] <- dying.ASD[, iS] * discounts;
    maturing.ASD[, iS] <- maturing.ASD[, iS] * discounts;
  }
 
  SSD <- apply(SASD, 2, sum)
  maturing.frac <- dying.frac <- numeric(num.stages);
  for(iS in 1:final.stage) {
    dying.frac[iS] <- sum(dying.ASD[,iS])/SSD[iS]
  }

  for(iS in 1:final.stage) {
    maturing.frac[iS] <- sum(maturing.ASD[,iS]) / SSD[iS]
  }
  total <- sum(SSD);
  SASD <- SASD/total;
  SSD <- SSD/total;
  

  as.VD.SASD(list(SASD = SASD,
                  SSD = SSD,
                  maturing.frac = maturing.frac,
                  dying.frac = dying.frac,
                  age.vector = age.vector))

}

VD.deriv.growth.rate <- function(VDM, r, dev.table, fecundity.by.age, stage = NULL, death = NULL, fecundity = NULL, param = NULL, force.numerical = FALSE, make.discrete = FALSE, delta = NULL, F = NULL) {
  denom <- VD.deriv.euler.dr(r, fecundity.by.age);
  numer <- VD.deriv.euler(VDM, r, dev.table, fecundity.by.age, stage, death, fecundity, param, force.numerical, make.discrete, delta, F)
  -numer/denom
}

VD.deriv.euler <- function(VDM, r, dev.table, fecundity.by.age, stage = NULL, death = NULL, fecundity = NULL, param = NULL, force.numerical = FALSE, make.discrete = FALSE, delta = NULL, F = NULL) {
  ## must get non-age-total dev times with contin if needed.
  
  use.stage <- !is.null(stage)
  use.death <- !is.null(death)
  use.fecundity <- !is.null(fecundity)
  if(sum(use.stage, use.death, use.fecundity) != 1) stop ("Error, must specify exacty one of stage, death, or fecundity");

  if(is.null(param)) {stop("Error: Must specify a param.");}
  if(!is.character(param)) {stop("Error: param must be a string.");}
  
  derivFunName <- paste("DerivToP.ratio.", param, sep = "");
  
  if(use.stage) {
    if(!is.numeric(stage)) {stop("Error: stage must be numerical.");}
    tableName <- "dev.table.contin" #to avoid total.age #if(!is.null(dev.table[["dev.table.contin"]])) "dev.table.contin" else "dev.table"
    if(is.null(dev.table[[tableName]])) {stop("Error: you must have dev.table.contin in dev.table");}
    if(stage < 0 | stage > num.stages(VDM)) {stop("Error: stage is out of range.");}
    
    if(stage.independent(VDM)[stage]) {
      if(make.discrete) {
        deriv.w.d.param <- VD.dist.deriv(dev.table[[tableName]][, stage], marginal.durations(VDM)[[stage]], "pdf", param, make.discrete = TRUE, force.numerical = force.numerical, relative = TRUE);
      } else {
        if(tableName == "dev.table") writeLines("Warning: This is only correct if the stage distribution is inherently discrete.")
        deriv.w.d.param <- VD.dist.deriv(dev.table[[tableName]][, stage], marginal.durations(VDM)[[stage]], "pdf", param, make.discrete = FALSE, force.numerical = force.numerical, relative = TRUE);
      }
    } else {
      ## now force.numerical can be a vector of two logicals, first for the overall deriv, the other for the dF(t)/d theta.
      #deriv.w.d.param <- VD.copula.deriv(dev.table[[tableName]][, stage], VDM, dev.table$Z, stage = stage, param = param, force.numerical = force.numerical, relative = TRUE);
      stop("Error: derivatives for copula models are not implemented yet.");
    }
  }

  if(use.death) {
    if(make.discrete) {writeLines("Warning: make.discrete does not do anything for a death derivative.");}
    if(!is.numeric(death)) {stop("Error: death must give a stage index.");}
    if(is.null(dev.table[["dev.table.contin"]])) {stop("Error: you must have dev.table.contin in dev.table");}
    values <- ceiling(dev.table$dev.table.contin[, death]) ## use dev.table.contin to avoid cumulative ages (total.age)
    deriv.w.d.param <- VD.dist.deriv(values, marginal.death.times(VDM)[[death]], "cdf", param, force.numerical = force.numerical, relative = TRUE, new.params = list(lower.tail = FALSE));
  }

  if(use.fecundity) {
    return(euler(r, calc.average.surv.rep.by.age(dev.table, F, use.shared.weight = TRUE, deriv.param = param)))
  } else {
    return(euler(r, calc.average.surv.rep.by.age(dev.table, F, weights = deriv.w.d.param, use.shared.weight = TRUE)))
  }
}


lfraction.maturing.at.SASD.independent <- function(VD.dist.dev, VD.dist.surv, r = 0, subtract.val = 0, max.age = NULL, VD.dist.dev.replace = NULL) {
  if(is.null(max.age)) max.age <- ceiling( min(VD.dist.calc(0.9999, VD.dist.surv, "quantile"),
       VD.dist.calc(0.9999, VD.dist.dev, "quantile") ) )

  if(max.age < 1) return(NA);
  age.vector <- 1:max.age;
  if(is.null(VD.dist.dev.replace)) {
    g.dev <- VD.dist.calc(age.vector, VD.dist.dev, "pdf", make.discrete = TRUE);
  } else {
    g.dev <- VD.dist.calc(age.vector, VD.dist.dev, "pdf", make.discrete = TRUE, new.params = VD.dist.dev.replace);
  }
  h.dev <- c(1, 1 - cumsum(g.dev)[1:(max.age-1)]); # this is probability that maturation is greater than or equal to a
  h.surv <- c(1, 1 - VD.dist.calc(age.vector, VD.dist.surv, "cdf", make.discrete = TRUE)[1:(max.age-1)]); #prob death is greater than or equal to a
  
  growth.discounts <- exp(-r * (age.vector-1));
  numerator <- sum( (growth.discounts[1:(max.age-1)] * h.surv[2:max.age] * g.dev[1:(max.age-1)]) )
  denominator <- sum( growth.discounts[1:(max.age-1)] * h.surv[2:max.age] * h.dev[1:(max.age-1)])

  log(numerator) - log(denominator) - subtract.val
}

VD.solve.fraction.maturing.at.SASD.independent <- function(VD.dist.dev, VD.dist.surv, r, fraction.maturing, param, range, max.age = NULL, details = FALSE) {
  lfraction.maturing <- log(fraction.maturing)
  replace.param <- list();
  replace.param[[param]] <- 0;
  objective <- function(v) {
    replace.param[[1]] <- v;
    lfraction.maturing.at.SASD.independent(VD.dist.dev, VD.dist.surv, r, lfraction.maturing, VD.dist.dev.replace = replace.param, max.age = max.age);
  }  
  ans <- uniroot(objective, range);
  if(details) ans else ans$root
}

lfraction.maturing.at.SASD.mc.dev.only <- function(VDM, VD.dist.dev, r, marg.cum, h.surv, age.prev, replace.param, age.vector, subtract.val = 0 ) {
	
  dev.time.contin <- VD.dist.calc(marg.cum, VD.dist.dev, "quantile", new.params = replace.param);
  dev.time <- ceiling(dev.time.contin);
  dev.age <- dev.time + age.prev;
  h.surv <- c(h.surv, 0); ## should never need the 0, but at least it will prevent an error if needed.
  maturing <- surviving <- rep(0, length(age.vector)); 
  for(a in age.vector) {
    ai <- a;
    bool <- dev.age >= a & age.prev < a;
    bool.now <- dev.age == a;
    age.in.stage <- a - age.prev;
    maturing[ai] <-  sum(h.surv[ age.in.stage[bool.now] + 1 ]);
    surviving[ai] <- sum(h.surv[ age.in.stage[bool] + 1 ]);
    bool.keep <- dev.age >= a;
    dev.age <- dev.age[bool.keep];
    age.prev <- age.prev[bool.keep];
  }
  
  discounts <- exp(-r * (age.vector-1))
  frac.mat <- sum(discounts * maturing) / sum(discounts * surviving);
  log(frac.mat) - subtract.val
}


lfraction.maturing.at.SASD.mc <- function(VDM, VD.dist.dev, r, marg.cum, death.time, age.prev, replace.param, max.age, subtract.val = 0 ) {
  dev.time.contin <- VD.dist.calc(marg.cum, VD.dist.dev, "quantile", new.params = replace.param);
  dev.time <- ceiling(dev.time.contin);
#  surv.bool <- death.time > dev.time;
#  die.bool <- !surv.bool;
  dev.age <- dev.time + age.prev;
  die.age <- death.time + age.prev;
  age.vector <- 1:max.age;
  maturing <- surviving <- rep(0, max.age)
  for(a in age.vector) {
    ai <- a;
    bool <- die.age > a & dev.age >= a;
    maturing[ai] <- sum(dev.age == a & bool);
    surviving[ai] <- sum(bool & age.prev < a);
    die.age <- die.age[bool];
    dev.age <- dev.age[bool];
    age.prev <- age.prev[bool];
  }
  
  discounts <- exp(-r * (age.vector-1))
  frac.mat <- sum(discounts * maturing) / sum(discounts * surviving);
  log(frac.mat) - subtract.val
}

VD.solve.fraction.maturing.at.SASD.mc <- function(VDM, iStage = -1, VD.dist.dev, VD.dist.surv, r, fraction.maturing, param, range,
                                               n = 10000, max.age = NULL, details = FALSE, mc.dev.only = FALSE) {

  lfraction.maturing <- log(fraction.maturing);
  replace.param <- list();
  replace.param[[param]] <- 0;
  if(iStage > 1) {
    VDS <- VD.run(VDM, 1:(iStage-1))
  } else { 
    VDS <- VD.sim(VDM);
  }
  dev.table <- compile.dev.table(VDS, total.age = TRUE, survivors = TRUE);
  age.prev <- rep(dev.table$dev.table[, iStage - 1], length.out = n); # count on correspondence to rep(...) below
  
  if(iStage > 1) {
    conditionalInfo <- VD.calc.conditional.info(VDM, VDS, iStage-1);
    conditionalMeans <- rep(conditionalInfo$conditional.means, length.out = n);
    conditionalSD <- conditionalInfo$conditional.sd;
  } else {
    conditionalMeans <- rep(0, n);
    conditionalSD <- 1;
  }
  cond.norm <- rnorm(n, conditionalMeans, conditionalSD);
  marg.cum <- pnorm(cond.norm);
  if(missing(VD.dist.surv)) VD.dist.surv <- marginal.death.times(VDM)[[iStage]];

  if(mc.dev.only) {
    age.vector <- 1:max.age;
    h.surv <- c(1, 1 - VD.dist.calc(age.vector, VD.dist.surv, "cdf", make.discrete = TRUE)[1:(max.age)]); #prob death is greater than or equal to a
    objective <- function(v) {
      replace.param[[1]] <- v;
      lfraction.maturing.at.SASD.mc.dev.only(VDM, VD.dist.dev, r,  marg.cum, h.surv, age.prev, replace.param, age.vector, lfraction.maturing)
    }
  } else {
    death.time <- ceiling(VD.dist.calc(n, VD.dist.surv, "random"));
    objective <- function(v) {
      replace.param[[1]] <- v;
      lfraction.maturing.at.SASD.mc(VDM, VD.dist.dev, r,  marg.cum, death.time, age.prev, replace.param, max.age, lfraction.maturing)
    }
  }

  ans <- try(uniroot(objective, interval = range));
  if(inherits(ans, "try-error")) return(NA)
  if(details) ans else ans$root
}

calc.copula.correlations <- function(dist1, dist2, rho) {

  gauss.cov <- matrix(c(1, rho, rho, 1), c(2,2));
  VDM <- VD.model(2, marginal.durations = list(dist1, dist2), marginal.death.times = list(VD.dist("infinite"), VD.dist("infinite")), gauss.cov = gauss.cov)
  controls(VDM)$dead.target <- 0;
  VDS <- VD.run(VDM);
  dev.table <- compile.dev.table(VDS, total.age = FALSE, include.contin = TRUE, include.Z = TRUE);
  corZ = cor(dev.table$dev.table.Z[,1], dev.table$dev.table.Z[,2]);
  cordtI = cor(dev.table$dev.table[,1], dev.table$dev.table[,2]);
  cordtC = cor(dev.table$dev.table.contin[,1], dev.table$dev.table.contin[,2]);
  list(cor.norm = corZ, cor.dev.time = cordtI, cor.dev.time.contin = cordtC);
}
